
__version__='$Revision: 1.3.2.1 $'[11:-2]

import os, string, base64, BTree, time
import MySQLdb

from ZODB import POSException, BaseStorage, utils
from ZODB.TimeStamp import TimeStamp

from binascii import b2a_base64, a2b_base64

def b2a(s): 
    return s and b2a_base64(s)[:-1] or s

def a2b(s):
    return s and a2b_base64(s) or s

logfile = open('RDBS.log', 'w')
def log(s):
    logfile.write(s)
    logfile.write('\n')
    logfile.flush()
    
class CorruptedDataError(POSException.StorageError):
    """Corrupted database storage.
    """

class MySQLStorage(BaseStorage.BaseStorage):
        
    def __init__(self, conn, name='MySQL Storage'):

        BaseStorage.BaseStorage.__init__(self, name)

        import pdb
#        pdb.set_trace()
        
        self.conn=conn
        
        # We use a BTree because the items are sorted!
        self._data=BTree.BTree()
        self._index={}
        self._vindex={}
        self._tindex=[]
        self._current=None
        self._read_index()
        keys=self._index.keys()
        if keys: self._oid=max(keys)

    def _escape(self, s):
        return MySQLdb.escape_string(s)

    def _read_index(self):
        """Load the database index"""
        import pdb
#        pdb.set_trace()
        data=self._data
        index=self._index
        vindex=self._vindex
        cur=self.conn.cursor()
        cur.execute(self._query_transaction_index)
        r=cur.fetchone()
        while r:
            serial,status,u,d,e=r
            serial=a2b(serial)
            data[serial]=status,u,d,e,[]
            r=cur.fetchone()
        cur.execute(self._query_data_index)
        r=cur.fetchone()
        while r:
            oid,serial,version=r
            oid=a2b(oid)
            serial=a2b(serial)
            index[oid]=serial
            if version:
                v=vindex.get(version, None)
                if v is None: v=vindex[version]={}
                v[oid]=serial
            data[serial][-1].append(oid)
            r=cur.fetchone()

    def __len__(self):
        return len(self._index)
        
    def getSize(self):
        """TBD"""
        return BaseStorage.BaseStorage.getSize(self)
        
    def supportsUndo(self): return 1
    def supportsVersions(self): return 1

    def load(self, oid, version):

        import pdb
#        pdb.set_trace()
#        log('load:\noid: %s' % repr(oid))
        
        self._lock_acquire()
        try:
            escape=self._escape
            serial=self._index[oid]
            cur=self.conn.cursor()
            cur.execute(self._query_load_data % (
                escape(b2a(oid)),
                escape(b2a(serial))))
            pre,oversion,nv,data = cur.fetchone()
#            log('loaded:\noid: %s serial: %s version: %s nv: %s data: %s' % (repr(oid), repr(serial), oversion, repr(nv), repr(data)))

            if oversion:
                if oversion != version:
                    nv=a2b(nv)
                    if nv:
                        serial=nv
                        cur.execute(self._query_load_data % (
                            escape(b2a(oid)),
                            escape(b2a(serial))))
                        pre,version,nv,data=cur.fetchone()
                    else: raise KeyError, oid
                    
            if not data: raise KeyError, oid
            return data, serial
        finally: self._lock_release()
                    
    def modifiedInVersion(self, oid):
        self._lock_acquire()
        try:
            try:
                escape=self._escape
                serial=self._index[oid]
                cur=self.conn.cursor()
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data = cur.fetchone()
                if version: return version
                return ''
            except: return ''
        finally: self._lock_release()

    def store(self, oid, serial, data, version, transaction):

        import pdb
#        pdb.set_trace()
#        log('store:\noid: %s serial: %s data: %s' % (repr(oid), repr(serial), repr(data)))
        
        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            if self._index.has_key(oid):
                old=self._index[oid]
            else:
                old=''

            nv=''
            if old:
                escape=self._escape
                cur=self.conn.cursor()
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(old))))
                pre,oversion,nv,odata = cur.fetchone()
                if oversion:
                    if oversion != version:
                        raise POSException.VersionLockError, oid
                    nv=a2b(nv)
                else:
                    nv=old
                if serial != old: raise POSException.ConflictError
                
            serial=self._serial
            self._store(oid,serial,old,version,nv,data)
            self._tindex.append((oid,serial,version))
#            log('stored:\noid: %s serial: %s pre: %s version: %s nv: %s data:%s' % (repr(oid), repr(serial), repr(old), version, repr(nv), repr(data)))
        finally: self._lock_release()

        return serial
        
    def _store(self, oid, serial, pre, version, nv, data):
        escape=self._escape
        cur=self.conn.cursor()
        cur.execute(self._query_insert_data % (
            escape(b2a(oid)),
            escape(b2a(serial)),
            'n',
            escape(b2a(pre)), 
            escape(version),
            escape(b2a(nv)),
            escape(data)
        ))
        if self._current is None: self._current=serial
        elif self._current != serial: raise POSException.ConflictError
        
    def _clear_temp(self):
        cur=self.conn.cursor()
        cur.execute(self._query_clear_temp)
        self._current=None
        self._tindex=[]
        
    def _finish(self, tid, user, desc, ext):

        import pdb
#        pdb.set_trace()
#        log('commit:')

        if self._current != tid: raise POSException.ConflictError
        
        escape=self._escape
        cur=self.conn.cursor()
        cur.execute(self._query_insert_transaction % (
            escape(b2a(tid)),
            escape(user),
            escape(desc),
            escape(ext)))
        cur.execute(self._query_commit_data % escape(b2a(tid)))
        self._current=None
        index=self._index
        vindex=self._vindex
        cur=self.conn.cursor()
        records=[]
        for oid,serial,version in self._tindex:
            old=index.get(oid, None)
            if old is not None:
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(old))))
                oldv=cur.fetchone()[1]
                if oldv:
                    v=vindex[oldv]
                    del v[oid]
                    if not v: del vindex[oldv]
                    
            index[oid]=serial
            records.append(oid)
            
            if version:
                v=vindex.get(version, None)
                if v is None: v=vindex[version]={}
                v[oid]=serial
            
        self._data[tid]='c',user,desc,ext,records

    def undo(self, tid):
        import pdb
#        pdb.set_trace()
        self._lock_acquire()
        try:
            _data=self._data
            tid=base64.decodestring(tid+'\n')
            try: t=_data[tid][4]
            except KeyError:
                raise UndoError, 'Invalid undo transaction id'

            index=self._index
            vindex=self._vindex
            escape=self._escape
            cur=self.conn.cursor()
            d=[]
            for oid in t:
                if index[oid] != tid:
                    raise POSException.UndoError, 'Undoable transaction'
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(tid))))
                pre,version,nv,data = cur.fetchone()
                pre=a2b(pre)
                nv=a2b(nv)
                d.append((oid,tid,pre,version,nv))
                
            oids=[]
            for oid,serial,pre,version,nv in d:
                cur.execute(self._query_undo_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                if pre:
                    index[oid]=pre
                    oids.append(oid)

                    # Delete old version data
                    if version:
                        v=vindex.get(version, None)
                        if v: del v[oid]

                    # Add new version data (from pre):
                    cur.execute(self._query_load_data % (
                        escape(b2a(oid)),
                        escape(b2a(pre))))
                    prepre,version,nv,data = cur.fetchone()
                    if version:
                        v=vindex.get(version, None)
                        if v is None: v=vindex[version]={}
                        v[oid]=pre
                        
                else:
                    del index[oid]
                    if version:
                        v=vindex.get(version, None)
                        if v: del v[oid]
                        if not v: del vindex[version]
                        
            cur.execute(self._query_undo_transaction % escape(b2a(tid)))
            del _data[tid]
            return oids
            
        finally: self._lock_release()

    def undoLog(self, first, last, filter=None):
        import pdb
#        pdb.set_trace()
        self._lock_acquire()
        try:
            _data=self._data.items()
            pos=len(_data)
            encode=base64.encodestring
            l=[]
            append=l.append
            i=0
            while i < last and pos:
                pos=pos-1
                if i < first:
                    i = i+1
                    continue
                tid, (s, u, d, e, t) = _data[pos]
                if s=='p': continue
                d={'id': encode(tid)[:-1],
                   'time': TimeStamp(tid).timeTime(),
                   'user_name': u, 
                   'description': d}
                if e:
                    d.update(loads(e))

                if filter is None or filter(d):
                    append(d)
                    i=i+1
            return l
        finally: self._lock_release()

    def versionEmpty(self, version):
        return not self._vindex.get(version, None)

    def versions(self, max=None):
        r=[]
        a=r.append
        for version in self._vindex.keys():
            if self.versionEmpty(version): continue
            a(version)
            if max and len(r) >= max: return r

        return r
        
    def commitVersion(self, src, dest, transaction):
        import pdb
#        pdb.set_trace()

        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)
        
        self._lock_acquire()
        try:
            v=self._vindex.get(src, None)
            if v is None: return
            
            tindex=self._tindex
            escape=self._escape
            cur=self.conn.cursor()
            _store=self._store
            oids=[]
            for oid, serial in v.items():
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data=cur.fetchone()
                oids.append(oid)
                pre=serial
                serial=self._serial
                _store(oid,serial,pre,'','',data)
                tindex.append((oid, serial, None))

            return oids

        finally: self._lock_release()

    def abortVersion(self, src, transaction):
        import pdb
#        pdb.set_trace()

        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)
        
        self._lock_acquire()
        try:
            v=self._vindex.get(src, None)
            if not v: return
            
            tindex=self._tindex
            escape=self._escape
            cur=self.conn.cursor()
            _store=self._store
            oids=[]
            for oid, serial in v.items():
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data=cur.fetchone()
                nv=a2b(nv)
                if nv:
                    oids.append(oid)
                    cur.execute(self._query_load_data % (
                        escape(b2a(oid)),
                        escape(b2a(nv))))
                    pre,version,nv,data=cur.fetchone()
                    pre=serial
                    serial=self._serial
                    _store(oid,serial,pre,'','',data)
                    tindex.append((oid,serial,None))
                else:
                    # effectively, delete the thing
                    pre=serial
                    serial=self._serial
                    _store(oid,serial,pre,'','','')
                    tindex.append((oid,serial,None))

            return oids

        finally: self._lock_release()
        
    def _build_indexes(self, stop='\377\377\377\377\377\377\377\377'):
        # Rebuild index structures from transaction data
        index={}
        vindex={}
        _data=self._data
        escape=self._escape
        cur=self.conn.cursor()
        for tid, (s, u, d, e, t) in _data.items():
            if tid >= stop: break
            for oid in t:
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(tid))))
                pre,version,nv,data = cur.fetchone()
                old=index.get(oid, None)
                if old is not None:
                    cur.execute(self._query_load_data % (
                        escape(b2a(oid)),
                        escape(b2a(old))))
                    oldversion=cur.fetchone()[1]
                    if oldversion:
                        v=vindex[oldversion]
                        del v[oid]
                        if not v: del vindex[oldversion]
                index[oid]=tid

                if version:
                    v=vindex.get(version, None)
                    if v is None: v=vindex[version]={}
                    vindex[version][oid]=tid
                    
        return index,vindex

    def pack(self, t, referencesf):

        import pdb
#        pdb.set_trace()
        
        self._lock_acquire()
        try:

            stop=`apply(TimeStamp, time.gmtime(t)[:5]+(t%60,))`
            _data=self._data
    
            # Build indexes up to the pack time:
            index,vindex = self._build_indexes(stop)
    
            # Now build an index of *only* those objects reachable
            # from the root.
            rootl=['\0\0\0\0\0\0\0\0']
            pop=rootl.pop
            pindex={}
            referenced=pindex.has_key
            escape=self._escape
            cur=self.conn.cursor()
            while rootl:
                oid=pop()
                if referenced(oid): continue
    
                # Scan non-version pickle for references
                serial=index[oid]
                pindex[oid]=serial
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data = cur.fetchone()
                nv=a2b(nv)
                referencesf(data, rootl)
                if version:
                    if nv:
                        cur.execute(self._query_load_data % (
                            escape(b2a(oid)),
                            escape(b2a(nv))))
                        pre,version,nv,data = cur.fetchone()
                        referencesf(data, rootl)
                
            # Now we're ready to do the actual packing.
            # We'll simply edit the transaction data in place.
            # We'll defer deleting transactions till the end
            # to avoid messing up the BTree items.
            deleted=[]
            for tid, (s, u, d, e, t) in _data.items():
                if tid >= stop: break
                o=[]
                for oid in t:
                    c=pindex.get(oid)
                    if c is None:
                        # GC this record, no longer referenced
                        continue
                    elif c != tid:
                        # This record is not the indexed record,
                        # so it may not be current. Let's see.
                        cur.execute(self._query_load_data % (
                            escape(b2a(oid)),
                            escape(b2a(tid))))
                        pre,version,nv,data = cur.fetchone()
                        if version:
                            # Version record are current *only* if they
                            # are indexed
                            continue 
                        else:
                            # OK, this isn't a version record, so it may be the
                            # non-version record for the indexed record.
                            cur.execute(self._query_load_data % (
                                escape(b2a(oid)),
                                escape(b2a(c))))
                            pre,version,nv,data = cur.fetchone()
                            if version:
                                nv=a2b(nv)
                                if nv != tid:
                                    # This record is not the non-version
                                    # record for the indexed record
                                    continue
                            else:
                                # The indexed record is not a version record,
                                # so this record can not be the non-version
                                # record for it.
                                continue
                        
                    o.append(oid)
                    
                if o:
                    if len(o) != len(t):
                        for oid in t:
                            if oid not in o:
                                cur.execute(self._query_delete_data % (
                                    escape(b2a(oid)),
                                    escape(b2a(tid))))
                        _data[tid]='p',u,d,e,o # Reset data
                    else:
                        _data[tid]='p',u,d,e,t
                    cur.execute(self._query_pack_transaction % 
                        escape(b2a(tid)))
                else:
                    deleted.append((tid,t))
    
            # Now delete empty transactions
            for tid,t in deleted: 
                for oid in t:
                    cur.execute(self._query_delete_data % (
                        escape(b2a(oid)),
                        escape(b2a(tid))))
                cur.execute(self._query_delete_transaction % escape(b2a(tid)))
                del _data[tid]
            
            # Now reset previous pointers for "current" records:
            for oid,serial in pindex.items():
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data = cur.fetchone()
                if version:
                    cur.execute(self._query_pack_data % escape(nv))
                cur.execute(self._query_pack_data % escape(b2a(serial)))

            pindex=None
    
            # Finally, rebuild indexes from transaction data:
            self._index,self._vindex = self._build_indexes()

        finally: self._lock_release()

    def _splat(self):
        """Spit out a string showing state.
        """
        import pdb
#        pdb.set_trace()
        o=[]
        o.append('Transactions:')
        cur=self.conn.cursor()
        data=self._data
        index=self._index
        escape=self._escape
        for serial, (s, u, d, e, t) in data.items():
            o.append('  %s %s' % (TimeStamp(serial), `(s, u, d, e)`))
            for oid in t:
                cur.execute(self._query_load_data % (
                    escape(b2a(oid)),
                    escape(b2a(serial))))
                pre,version,nv,data = cur.fetchone()
                pre=a2b(pre)
                nv=a2b(nv)
                o.append('    '+`(
                    utils.u64(oid),
                    str(TimeStamp(serial)),
                    pre and str(TimeStamp(pre)) or pre,
                    version,
                    nv and str(TimeStamp(nv)) or nv,
                    data)`)

        o.append('Index:')
        keys=index.keys()
        keys.sort()
        for oid in keys:
            serial=index[oid]
            o.append('  %s: %s' % (utils.u64(oid),TimeStamp(serial)))

        o.append('\nVersion Index:')
        items=self._vindex.items()
        items.sort()
        for version, v in items:
            o.append('  '+version)
            vitems=v.items()
            vitems.sort()
            for oid, serial in vitems:
                o.append('    %s: %s' % (utils.u64(oid), TimeStamp(serial)))
                
        return string.join(o,'\n')
        
    _query_data_index="""
        select z_oid,z_serial,z_version
        from zodb_data
        where z_status='c';"""
    _query_transaction_index="""
        select z_serial,z_status,z_user,z_desc,z_ext
        from zodb_trans
        where z_status!='u';"""
    _query_load_data="""
        select z_pre,z_version,z_nv,z_data
        from zodb_data
        where z_oid='%s' and z_serial='%s';"""
    _query_insert_data="""
        insert into zodb_data 
        (z_oid, z_serial, z_status, z_pre, z_version, z_nv, z_data)
        values ('%s', '%s', '%s', '%s', '%s', '%s', '%s');"""
    _query_insert_transaction="""
        insert into zodb_trans
        (z_serial,z_status,z_user,z_desc,z_ext)
        values ('%s', 'c', '%s', '%s', '%s');"""
    _query_commit_data="""
        update zodb_data
        set z_status='c'
        where z_serial='%s';"""
    _query_clear_temp="""
        delete from zodb_data
        where z_status='n';"""
    _query_undo_data="""
        update zodb_data
        set z_status='u'
        where z_oid='%s' and z_serial='%s';"""
    _query_undo_transaction="""
        update zodb_trans
        set z_status='u'
        where z_serial='%s';"""
    _query_pack_data=r"""
        update zodb_data
        set z_pre=''
        where z_serial='%s';"""
    _query_pack_transaction="""
        update zodb_trans
        set z_status='p'
        where z_serial='%s';"""
    _query_delete_data="""
        delete from zodb_data
        where (z_oid='%s' and z_serial='%s') or z_status='u';"""
    _query_delete_transaction="""
        delete from zodb_trans
        where z_serial='%s';"""

